/*
 * http_post: A small library for POST data retrieval from CGI.
 * (c) 2007 Avelino Herrera Morales
 * M.H.P. Sistemas de Control S.L.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "http_post.h"

char delim[128];
int delim_size = 0;
char buffer[4096];
char fail_buffer[256];
char variable_name[80];
char file_name[256];
char trash[80];
char cache[2048];
int cache_index = 0;

#define  READING_CONTENT_DISPOSITION               0
#define  READING_HEADER_BEFORE_DATA                1
#define  READING_HEADER_BEFORE_VALUE               2
#define  READING_DATA                              3
#define  READING_VALUE                             4
#define  READING_BLANK_BEFORE_CONTENT_DISPOSITION  5

int reading = READING_BLANK_BEFORE_CONTENT_DISPOSITION;
FILE *dump_file;
char upload_path_prefix[256] = "/";
char upload_path[256];
unsigned char *dump_buffer_ptr;


post_data *search_post_data(post_data *root, int numvars, char *name) {
    int i;
    for (i = 0; i < numvars; i++) {
        if (strcmp(root[i].name, name) == 0)
            return &root[i];
    }
    return NULL;
}


void init_emit_char(post_data *pdata) {
    reading = READING_BLANK_BEFORE_CONTENT_DISPOSITION;
    cache_index = 0;
    pdata->what = WHAT_UNDEFINED;
    if (dump_file != NULL) {
        fclose(dump_file);
        dump_file = NULL;
    }
}

void done_emit_char(post_data *pdata) {
    if (reading == READING_VALUE) {
        cache[cache_index] = 0;
        // printf("done_emit_char:     value='%s'\n", cache);
        strcpy(pdata->un.value, cache);
        pdata->what = WHAT_VALUE;
    }
    else if (reading == READING_DATA) {
        // printf("done_emit_char:     file created\n");
        pdata->what = WHAT_FILE;
        if (dump_file != NULL) {
            fclose(dump_file);
            dump_file = NULL;
        }
    }
    else {
        pdata->what = WHAT_UNDEFINED;
    }
    reading = READING_BLANK_BEFORE_CONTENT_DISPOSITION;
    cache_index = 0;
}

void emit_char(char c, post_data *pdata) {
    int p, r;

    /*
    if (isgraph(c))
        printf("emit_char(%c) reading=%d\n", c, reading);
    else
        printf("emit_char(0x%02X) reading=%d\n", (int) c, reading);
    */
    fflush(stdout);
    switch (reading) {
        case READING_CONTENT_DISPOSITION:
            cache[cache_index] = c;
            cache_index++;
            if (c == '\x0A') {
                cache[cache_index] = 0;
                sscanf(cache, "Content-Disposition: form-data; name=\"%[^\"]\"%n", variable_name, &p);
                strcpy(pdata->name, variable_name);
                /* printf("                r=%d\n", r); */
                if (cache[p] == ';') {
                    sscanf(&cache[p], "; filename=\"%[^\"]\"\x0D\x0A", file_name);
                    // printf("file_name='%s'\n", file_name);
                    if (upload_path_prefix[0] != 0) {
                        strcpy(upload_path, upload_path_prefix);
                        strcat(upload_path, file_name);
                        strcpy(pdata->un.filename, file_name);
                        if ((dump_file = fopen(upload_path, "wb")) == NULL) {
                            perror("open file");
                            exit(-1);
                        }
                    }
                    else
                        dump_file = NULL;
                    reading = READING_HEADER_BEFORE_DATA;
                }
                else {
                    sscanf(&cache[p], "\x0D\x0A");
                    reading = READING_HEADER_BEFORE_VALUE;
                }
                cache_index = 0;
            }
            break;
        case READING_HEADER_BEFORE_VALUE:
        case READING_HEADER_BEFORE_DATA:
            cache[cache_index] = c;
            cache_index++;
            if (c == '\x0A') {
                cache[cache_index] = 0;
                if (sscanf(cache, "%[^\x0D\x0A]\x0D\x0A", trash) == 1) {
                    //fprintf(stderr, "http_post: trash='%s'\n", trash);
                    if (upload_path_prefix[0] == 0) {
                        if (sscanf(trash, "Content-Length: %d", &(pdata->un.buffer.size)) == 1) {
                            pdata->un.buffer.ptr = (unsigned char *) malloc(pdata->un.buffer.size);
                            dump_buffer_ptr = pdata->un.buffer.ptr;
                        }
                    }
                    reading = reading;   /* again */
                }
                else {
                    sscanf(cache, "\x0D\x0A");
                    if (reading == READING_HEADER_BEFORE_VALUE)
                        reading = READING_VALUE;
                    else
                        reading = READING_DATA;
                }
                cache_index = 0;
            }
            break;
        case READING_BLANK_BEFORE_CONTENT_DISPOSITION:
            cache[cache_index] = c;
            cache_index++;
            if (c == '\x0A') {
                cache[cache_index] = 0;
                cache_index = 0;
                reading = READING_CONTENT_DISPOSITION;
            }
            break;
        case READING_DATA:
            if (upload_path_prefix[0] != 0) {
                if (dump_file != NULL)
                    fwrite(&c, 1, 1, dump_file);
            }
            else {
                *dump_buffer_ptr = (unsigned char) c;
                dump_buffer_ptr++;
            }
            break;
        case READING_VALUE:
            cache[cache_index] = c;
            cache_index++;
            /*
            if (c == '\x0A') {
                cache[cache_index] = 0;
                printf("value='%s'\n", cache);
            }
            */
            break;
        default:
            break;
    }
}

int get_next_token(char *delim, char *buffer, int *input_index, int input_size, post_data *pdata) {
    int fail_index = 0;
    int delim_index = 0;
    //int output_index = 0;
    int delim_length = strlen(delim);
    int c, i;

    init_emit_char(pdata);
    memset(fail_buffer, 0, 256);
    while (*input_index < input_size) {
        c = getchar();
        (*input_index)++;
        if (delim[delim_index] == (char) c) {
            fail_buffer[fail_index] = (char) c;
            fail_index++;
            delim_index++;
            if (delim_index == delim_length) {
                //buffer[output_index] = 0;
                done_emit_char(pdata);
                return 1;
            }
        }
        else {
            for (i = 0; i < fail_index; i++) {
                //buffer[output_index] = fail_buffer[i];
                emit_char(fail_buffer[i], pdata);
                //output_index++;
            }
            fail_index = 0;
            delim_index = 0;
            //buffer[output_index] = c;
            emit_char(c, pdata);
            //output_index++;
        }
    }
    //buffer[output_index] = 0;
    done_emit_char(pdata);
    return 0;
}


void get_post_data(post_data **ret, int *numvars) {
    char *content_length_str;
    int res;
    int input_index = 0;
    int input_size;
    int p;
    post_data pdata;

    *numvars = 0;
    *ret = NULL;
    content_length_str = getenv("CONTENT_LENGTH");
    if (content_length_str == NULL)
        return;
    input_size = atoi(content_length_str);
    if (input_size > 0) {
        scanf("%s%n", &delim[2], &p);
        delim[0] = 0x0D;
        delim[1] = 0x0A;
        delim_size = strlen(delim);
        input_size -= p;
        do {
            res = get_next_token(delim, buffer, &input_index, input_size, &pdata);
            if (pdata.what == WHAT_UNDEFINED)
                continue;
            (*numvars)++;
            if (ret == NULL)
                *ret = malloc(sizeof(post_data));
            else
                *ret = realloc(*ret, *numvars * sizeof(post_data));
            memcpy(&(*ret)[*numvars - 1], &pdata, sizeof(post_data));
        } while (res);
    }
}


int http_post_main_test(int argc, char *argv[]) {
    post_data *pdata;
    int numvars;
    int i;

    get_post_data(&pdata, &numvars);
    printf("Content-type: text/html\r\n\r\n");
    printf("<html><body><pre>");
    printf("%d vars\r\n", numvars);
    for (i = 0; i < numvars; i++) {
        printf("name[i]=%s\r\n", pdata[i].name);
        if (pdata[i].what == WHAT_FILE)
            printf("\tfilename=%s\r\n", pdata[i].un.filename);
        else
            printf("\tvalue=%s\r\n", pdata[i].un.value);
    }
    printf("</pre></body></html>");
    return 0;
}
